/*
 * EthernetUtilities.java
 *
 * Created on June 21, 2007, 9:11 AM
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package ModelDataLink;

import Utility.JNWBitSet ;
import Utility.JNWRandom;

/**Utilities for Ethernet frames.  Creates Etherent frames, extracts
 *addresses, and computes Ethernet backoff.
 * @version 1.100
 * @author dixonw
 *@see <a href="http://en.wikipedia.org/wiki/Ethernet">Ethernet</a>
 *@see <a href="http://en.wikipedia.org/wiki/Image:Ethernet_Type_II_Frame_format.svg">Ethernet Type II Frame format</a>
 *@see <a href="http://www.geocities.com/SiliconValley/Vista/8672/network/ethernet.html">Ethernet II versus IEEE 802.3 </a>
 *@see <a href="http://www.geocities.com/SiliconValley/Haven/4824/ethernet.html">Ethernet Page</a>
 *@see <a href="http://alevi.usc.edu/Html/datacom/sld024.htm">Ethernet Frame Format</a>
 */
public class EthernetUtilities extends EthernetState implements ModelConstants.General
{
   /**Get the random number generator*/
   static JNWRandom randomNumberGenerator = JNWRandom . getInstance ( ) ;

   /**The singleton vector containing all the links*/
   ModelData.JNWSingletonVectorOfLinks<ModelPhysical.Link> vectorOfLinks;

   /**The singleton vector containing all the links*/
   ModelData.JNWSingletonVectorOfInterfaces<ModelDataLink.InterfaceBaseClass> vectorOfInterfaces
      = ModelData.JNWSingletonVectorOfInterfaces.getInstance();

   /** Creates a new instance of EthernetUtilities */
   public EthernetUtilities ()
   {
   }

   /**The Ethernet preamble, with 7 bytes of "10101010" and a SDF (start
    * frame delimiter), 10101011" is used to synchronize the receiver.*/
   public static JNWBitSet ethernetPreamble = new JNWBitSet (
      "10101010 10101010 10101010 10101010 "
      + "10101010 10101010 10101010 10101011 " ) ;


   /**The Ethernet CRC-32.  CRC-32-IEEE 802.3
    * @see <a href="http://en.wikipedia.org/wiki/Cyclic_redundancy_check">CRC-32-IEEE 802.3</a>
    */
   private static JNWBitSet CRC_32 = new JNWBitSet (
      "1 00000100 11000001 00001101 10110111" ) ;


   /**Adds the Ethernet header and checksum at the end of an Ethernet frame.
    *The method does <b>NOT</b> pad the data field, if necessary, to the
    * minimum length of 46 bytes. It does <b>NOT</b> fragment the Ethernet
    * frame if the frame length exceed 1500 bytes.
    * <p>This method reads the 32-bits source and destination addresses in
    * IP header of the format 0.0.0.N,  and it creates an 48-bit
    * MAC address of the format 00-00-00-00-00-N.*/
   public static JNWBitSet AddEthernetHeaderAndTrailer ( JNWBitSet packetIn )
   { // First clone the packet
      JNWBitSet ethernetFrame = packetIn.clone ( ) ;
      // Pad the data if less than 46 bytes long
      if ( ethernetFrame .getJNWLength () < 46 * BITS_PER_BYTE )
      { // Simply shift the data up as necessary
 // This has been eliminated
 ////        ethernetFrame .shift ( 46 * 8 - ethernetFrame .getJNWLength ( ) ) ;
      }
      // Compute the frame length in bytes including the preamble
      int ethernetFrameLengthInBytes = ethernetFrame.getJNWLength () / BITS_PER_BYTE
         + ( 8 /*preamble length*/
         +   6 /*destination address length*/  + 6 /*source address length*/
         +   2 /*length length */              + 4 /* CRC length */)  ;
      JNWBitSet lengthBitSet = new JNWBitSet ( ethernetFrameLengthInBytes , 16  ) ;
      ethernetFrame . prepend ( lengthBitSet) ;
      // Compute the source MAC address
      long sourceAddress
         = ModelTransport . IPUtilities
            . GetSourceAddressFromIPHeader ( packetIn ) ;
      JNWBitSet sourceAddressBitSet = new JNWBitSet ( sourceAddress , 48 ) ;
      ethernetFrame . prepend ( sourceAddressBitSet ) ;
      // Compute the destination MAC address
      long destinationAddress
         = ModelTransport . IPUtilities
            . GetDestinationAddressFromIPHeader ( packetIn ) ;
      JNWBitSet destinationAddressBitSet = new JNWBitSet ( destinationAddress , 48  ) ;
      ethernetFrame . prepend ( destinationAddressBitSet ) ;
      // Comput the CRC over the frame
      ethernetFrame . appendFCS ( CRC_32 ) ;
      // Add the preamble
      ethernetFrame . prepend ( ethernetPreamble ) ;
      return  ethernetFrame ;
   }  // end of public static JNWBitSet AddHeaderAndFooter


   /**Returns the destination MAC address from the Ethernet frame or
    * just the Ethernet frame header.*/
   public static long GetDestinationMACAddress ( JNWBitSet ethernetFrameIn )
   {
      int fromIndex = ethernetFrameIn . getJNWLength ( ) -  ( 8 /*preamble length*/
         + 6 /*destination address length*/ ) * 8 ;
      return ethernetFrameIn .getSubset ( fromIndex , fromIndex + 48 )
         . convertJNWBitSetToLong ( ) ;
   }  // end of GetDestinationMACAddress


   /**Returns the source MAC address from the Ethernet frame or
    * just the Ethernet frame header.*/
   public static long GetSourceMACAddress ( JNWBitSet ethernetFrameIn )
   {
      int fromIndex = ethernetFrameIn . getJNWLength ( ) -  ( 8 /*preamble length*/
         + 6 /*destination address length*/ + 6 /*source address length*/ ) * BITS_PER_BYTE ;
      return ethernetFrameIn . getSubset ( fromIndex , fromIndex + 48 )
         . convertJNWBitSetToLong ( ) ;
   }  // end of GetSourceMACAddress


  /**Verifies the Ethernet CRC.  The FCS is not performed over the preamble.
   *returns a true if the CRC is correct.  Return a false if the CRC is
   *incorrect, indicating transmission errors.*/
   public static boolean verifyAtReceiver ( JNWBitSet ethernetFrameIn ) {
     JNWBitSet ethernetFrameInClone = ethernetFrameIn . clone ( ) ;
     ethernetFrameInClone . setJNWLength ( ethernetFrameInClone . getJNWLength ( ) -
        8 /*preamble length*/ * BITS_PER_BYTE ) ;
     return ethernetFrameInClone  . verifyFCS ( CRC_32 ) ;
   }


   /**Returns the destination address from the IP header contained in an Ethernet frame.*/
   public static long GetDestinationAddressFromIPHeader ( JNWBitSet ethernetFrameIn )
   {
      int fromIndex = ethernetFrameIn . getJNWLength ( ) -  ( 8 /*preamble length*/
         + 6 /*destination address length*/  + 6 /*source address length*/
         + 2 /*length*/ + 4 * 5 /*IP header length to destination address*/) * 8 ;
      return ethernetFrameIn .getSubset ( fromIndex , fromIndex + 32 )
         . convertJNWBitSetToLong ( ) ;
   }


   /**Returns the source address from the IP header contained in an Ethernet frame.*/
   public static long GetSourceAddressFromIPHeader ( JNWBitSet ethernetFrameIn )
   {
      int fromIndex = ethernetFrameIn . getJNWLength ( ) -  ( 8 /*preamble length*/
         + 6 /*destination address length*/  + 6 /*source address length*/
         + 2 /*length*/ + 4 * 4 /*IP header length to source address*/) * 8 ;
      return ethernetFrameIn .getSubset ( fromIndex , fromIndex + 32 )
         . convertJNWBitSetToLong ( ) ;
   }


   /**Returns the data from the Ethernet frame.  The data will everything except
    *the Ethernet header and the Ethernet CRC. This will include the IP header,
    *the UDP or TCP header, and the data in the packet.  See appendix E of the
    *Users Guide.*/
   public static JNWBitSet GetIPPacketFromEthernetFrame ( JNWBitSet ethernetFrameIn )
   {
      int ethernetHeaderLength = ( 8 /*preamble Length in bytes*/
         + 6 /*destination address length*/ + 6 /*source address length*/
         + 2 /*message Length in bytes*/) * 8;
      return ethernetFrameIn .getSubset ( 4 /*4 byte CRC*/* 8
         , ethernetFrameIn . getJNWLength ( ) - ethernetHeaderLength ) ;
   }

   /**Returns the time delay in Ticks between successive Ethernet
    * transmission attempts on an Ethernet bus.
    * <li>1<= transmissionAttemptNumber<=
    * maximumNumberOfTransmissionAttempts </li>*/
   public static long ComputeBackoff ( int transmissionAttemptNumber )
   {
      // THIS IS ADDED ONLY SO THAT THE JNWS AND THE ALGORITHM DESCRIPTION
      // IN PULLEN AGREE
      // slot_ticks is the number of ticks in each Ethernet slot, here 51.2 microseconds      
      int slot_ticks = ModelDataLink.EthernetState.ethernetSlotSizeInTicks ;
//   STUDENT FILLS THIS IN
//   STUDENT FILLS THIS IN
//   SLIGHT ADAPTION FROM PAGE 82 of PULLEN

//   Algorithm for method ComputeBackoff ( )
//   if max_backoff_slots is less than 1024,
//        double the value of the max_backoff_slots;
//   if backoff_count is greater than 16,
//        return value -1 indicating failure;
//   generate a random_number between 0 and 1, using the random number
//   seed associated with this interface;
//   calculate random_backoff_slots = random_number * max_backoff_slots;
//   return backoff_ticks = random_backoff_slots * slot_ticks ;


//   STUDENT FILLS THIS IN TO RETURN backoff_ticks
//   STUDENT FILLS THIS IN TO RETURN backoff_ticks
//   STUDENT FILLS THIS IN TO RETURN backoff_ticks
      return -1 ;
   }

   /**For testing only.*/
   public static void main_Testing_Only ( String [ ] args )
   {
      JNWBitSet data = new JNWBitSet ( ) ;
      JNWBitSet thirtyTwoZeros = new JNWBitSet ( "00000000 00000000 00000000 00000000") ;
////    JNWBitSet twentyFourZeros = new JNWBitSet (
      ////           "00000000 00000000 00000000") ;
      data . prepend ( thirtyTwoZeros ) ;
      data . prepend ( thirtyTwoZeros ) ;
      data . prepend ( thirtyTwoZeros ) ;
      JNWBitSet sourceAddress = new JNWBitSet ( ) ;
      JNWBitSet destinationAddress = new JNWBitSet ( ) ;
      sourceAddress . setJNWBitSetToLongAndTruncate ( 1515 , 32 ) ;
      destinationAddress . setJNWBitSetToLongAndTruncate ( 1717 , 32 ) ;
      data . append ( sourceAddress ) ;
      data . append ( destinationAddress ) ;
      data . append ( thirtyTwoZeros ) ;
      data . append ( thirtyTwoZeros ) ;
      data . append ( thirtyTwoZeros ) ;
      System.out.println ("The IP packet (" + ( data.getJNWLength () / 8 ) + " bytes) is : " + data.toString ( ) ) ;
      System.out.println ("THE source address is "
         + ModelTransport.IPUtilities.GetSourceAddressFromIPHeader ( data ) ) ;
      System.out.println ("THE destination address is "
         + ModelTransport.IPUtilities.GetDestinationAddressFromIPHeader ( data ) ) ;
      JNWBitSet ethernetFrame = new JNWBitSet ( ) ;
      ethernetFrame = AddEthernetHeaderAndTrailer ( data  ) ;
      System.out.println ("The ethernet frame is : " + ethernetFrame.toString ( ) ) ;
      int destinationAddressInt = ( int ) GetDestinationMACAddress ( ethernetFrame ) ;
      System.out.println ("The ethernet destination address is : " + destinationAddressInt ) ;
      int sourceAddressInt = ( int ) GetSourceMACAddress ( ethernetFrame ) ;
      System.out.println ("The ethernet source address is : " + sourceAddressInt ) ;
      int fromIndex = ethernetFrame . getJNWLength ( ) - ( 8 + 6 + 6 + 2 ) * 8 ;
      JNWBitSet lengthBitSet =  ethernetFrame .getSubset ( fromIndex , fromIndex + 16 ) ;
      System.out.println ("The ethernet frame length is : " + lengthBitSet.convertJNWBitSetToLong ()  ) ;
      JNWBitSet dataFromEthernetFrame = GetIPPacketFromEthernetFrame ( ethernetFrame ) ;
      System.out.println ("The ethernet data (" + ( dataFromEthernetFrame.getJNWLength () / 8 ) + " bytes) is : "
         + dataFromEthernetFrame . toString ()  ) ;
      for ( int i = 0 ; i < ethernetFrame . getJNWLength ( ) - 8 /*preamble length*/ * 8 + 1 ; i++ ) {
      ethernetFrame . flip ( i ) ;
      System.out.println ("The ethernet CRC is " + verifyAtReceiver ( ethernetFrame ) ) ;
      ethernetFrame . flip ( i ) ;
   }  // end of main


   } // end of Main

}  // end of class
