/*
 * JNWBuffer.java
 *
 * Created on June 1, 2007, 1:26 PM
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package ModelTransport;

import Utility.*;

/**Creates a buffer to accept TCP messages.  Arriving packets are ordered
 * by sequence number.  Duplicate packets are discarded.  The buffer updates
 *the connection acquisition number.  The MessageComplete method returns
 *a boolean saying whether the buffer contsins a complete messages.  The
 *poll method returns a JNWBitSet containing the complete message.
 * *<p><b>Versions:</b>
 *<li><b>Version 1.001</b> of 20071123 replaces the java.util.PriorityQueue
 * with an array, dataArray.</li>
 * <li><b>Version 1.200</b> of 20080821 corrects an error in method Poll().  Added
 * two methods private void processFINBit ( int i ) 
 * and private void resetFINBitVariables ( ) to set the FIN bit variables.  I also added
 * a field sequenceNumbersAlreadyProcessed which prevents the buffer from receiving
 * duplicate packets if the packet has already been precessed by the poll ( ) method.
 * </li>
 * @author dixonw
 * @version 1.200
 */
public class TCPDestinationBuffer implements GUIConstants.General , ModelConstants.General
{
   /**An Object array for storing sequence numbers, FIN bits, acknowledgement
    *numbers, and the data as a JNWBitSet.  The fields are:
    *<li>[ 0 ] a Long containing the packet sequence number,</li>
    *<li>[ 1 ] a Boolean containing the packet FIN bit,</li>
    *<li>[ 2 ] a Long containing the next expected sequence number, which
    * is the sequence number plus the data length, and</li>
    *<li>[ 3 ] a JNWBitSet containing the data.</li>*/
   Object [ ] [ ] dataArray = new Object [ 30 ] [ 4 ] ;

   /**An array of subscripts for the dataArray.  It becomes necessary
    *only to change the subscript array as packet locations change,
    *rather than changing the dataArray.*/
   int [ ] dataArrayRowSubscript = new int [ dataArray .length ] ;

   /**Maximum Long value. A large number of 11 digits*/
   Long longMaxValue
      = new Long ( ( long ) 1000 *  1000 *  1000 * 100 - 1 ) ;

   /**True if one or more TCP FIN bits has been received.*/
   boolean FINBitHasBeenReceived = false ;

   /**The sequence number of the <b>first</b> packet with the FIN bit set.*/
   long sequenceNumberOfFirstFINBit = MINUS_ONE ;


   /**Acknowledgement number for the buffer.  This is the next sequence
    * number that the buffer is expecting.  This number changes as new
    * TCP packets are received.  This number will have to
    * change if there is a real TCP "Handshake" to open the connection.
    * bufferAcknowledgementNumber is initialized in the constructor
    * of TCPState. */

   // Important (bo)
   long bufferAcknowledgementNumber  ;
   
   
   /**The sequence numbers that have already been processed by the 
    *poll method.  Retransmissions of packets that have already been 
    * processed are ignored by the add method.*/
   long sequenceNumbersAlreadyProcessed ; 


   /** Creates a new instance of TCPDestinationBuffer.*/
   public TCPDestinationBuffer (  )
   {  // Fill the dataArray with blank data
      for ( int i = 0 ; i < dataArray . length ; i++ )
      {
         dataArray [ i ] [ 0 ] = longMaxValue ;
         dataArray [ i ] [ 1 ] = Boolean . FALSE ;
         dataArray [ i ] [ 2 ] = longMaxValue ;
         dataArray [ i ] [ 3 ] = null ;
         dataArrayRowSubscript [ i ] = i ;
      }
      resetFINBitVariables ( ) ;    
   }  // end of TCPDestinationBuffer(  )


/**Adds a JNWBitSet IP packet to the TCP buffer in order of sequence number.
 *  It first tests to ensure that the IP packets is not already in the
 *queue.  If it is, the processing ends.  If not, the method updates
 *numberOfMessageBytesInQueue, firstSequenceNumberInQueue, and
 *lastByteInQueue. If the TCP packet contains the FIN bit,
 *FINBitHasBeenReceived  is set to true.*/
   public boolean add ( JNWBitSet IPPacketIn )
   {  long IPPacketSequenceAddress
         = ModelTransport . TCPUtilities
         . GetTCPSequenceNumberFromIPPacket ( IPPacketIn ) ;
      JNWBitSet data
         = ModelTransport . TCPUtilities . GetDataCloneFromTCP_IPPacket ( IPPacketIn ) ;
      for ( int i = 0 ; i < dataArray.length ; i++ )
      {  // Has this sequence number already been processes by method poll ( ) ?

         // Duplication detection (bo)
         if ( IPPacketSequenceAddress < sequenceNumbersAlreadyProcessed )
         {  // The sequence number has already been processesed
            return false ;
         }
         // Compare the sequence number and the dataArray entry
         int comparison = new java.lang.Long ( IPPacketSequenceAddress )
         . compareTo ( ( java.lang.Long ) dataArray [
            dataArrayRowSubscript [ i ] ] [ 0 ] )  ;
         if ( comparison == 0 )
         {  // The sequence number is already in the dataArray
            return false ;
         }
         // Is the sequence number greater than the dataArray entry, insert
         // the sequence number in the next row
         else if ( comparison < 0 )
         {  // Open the next row
            OpenRowNOfDataArray ( i ) ;
            // Enter the data.
            dataArray [ dataArrayRowSubscript [ i ]  ] [ 0 ]
               = new Long ( IPPacketSequenceAddress ) ;
            // has the FIN bit been received?
            if ( ModelTransport . TCPUtilities
               . GetTCPFINBitFromIPPacket ( IPPacketIn ) == true )
            {  //  The FIN bit has been received
               processFINBit ( i ) ;
            }  // else
            else
            {
               dataArray [ dataArrayRowSubscript [ i ]  ] [ 1 ] = Boolean . FALSE ;
            }
            // Compute the sequence number of the next expected packet
            Long nextSequenceNumber = new Long ( IPPacketSequenceAddress
            +  data . getJNWLength ( ) / BITS_PER_BYTE  ) ;           
            dataArray [ dataArrayRowSubscript [ i ] ] [ 2 ] = nextSequenceNumber ;
            // Save the packet
            dataArray [ dataArrayRowSubscript [ i ] ] [ 3 ] = data  ;

            // Does this packet addition increase the acknowledgement number?
            if ( bufferAcknowledgementNumber == IPPacketSequenceAddress )
            {
               bufferAcknowledgementNumber = nextSequenceNumber ;
               // But this could be a missing packet therefore update the
               // acknowledgement number may be different
               // Update the acknowledgement number
               UpdateAcknowledgementNumber ( ) ;
            }
            return true ;
         }  // end of else if ( comparison < 0 )
         // Otherwise continue up the array.
      }  // for ( int i = 0 ; i < dataArray.length ; i++ )

      return true ;
   }  // end of public void add ( Utilities.JNWBitSet bitSet )

   
/**Processes the FIN bit of the packet.  Updates the boolean
 FINBitHasBeenReceived when the first FIN bit is received.  Updates
 the sequenceNumberOfFirstFINBit to the sequence number of the
 <b>first</b> packet containing the FIN bit set.  Call this method ONLY
 when you know that the ith row contains a FIN bit set.*/
 private void processFINBit ( int i ) {
    // Is this the first FIN bit found?
    if ( FINBitHasBeenReceived == false ) {
       sequenceNumberOfFirstFINBit
          = ( ( Long ) dataArray [ dataArrayRowSubscript [ i ] ] [ 0 ] ) . longValue ( ) ;
    }
    // If the new sequence number of the FIN bit is LESS than
    // sequenceNumberOfFirstFINBit, update sequenceNumberOfFirstFINBit
    if ( new java.lang.Long ( sequenceNumberOfFirstFINBit )
       . compareTo ( ( java.lang.Long ) dataArray [
       dataArrayRowSubscript [ i ] ] [ 0 ] ) > 0 )  {
          sequenceNumberOfFirstFINBit
             = ( ( Long ) dataArray [ dataArrayRowSubscript [ i ] ] [ 0 ] ) . longValue ( ) ;
       }  // end of if ( new java.lang.Long ( sequenceNumberOfFirstFINBit )
    // Now set FINBitHasBeenReceived to true
    FINBitHasBeenReceived = true ;
    // Set the boolean in the dataArray
    dataArray [ dataArrayRowSubscript [ i ] ] [ 1 ] = Boolean . TRUE ;
}  // end of processFINBit


/**Resets the FIN bit varaibles FINBitHasBeenReceived and
 *  sequenceNumberOfFirstFINBit.  This is called from method poll ( ) 
 * after a message has been returned by the method.*/
 private void resetFINBitVariables ( ) {
    FINBitHasBeenReceived = false ;
    sequenceNumberOfFirstFINBit = MINUS_ONE ;
 }

   /**Returns a bit set containing the complete message and it resets the
    * TCPDestinationBuffer.  If no complete message is in the buffer, poll
    * return null.  If a complete message exists poll saves the message in
    * resultString, removes the message from dataArray and resursively
    * to find the next complete message, if it exists.  As the recursive
    * calls unwind the later messages are arred to resultString.*/
   public  String  poll (  )
   {  // Has a FIN bit been received?  If not
      // a complete message cannot be in the buffer
      // If the FIN bit has not been received
      // or the first message is incomplete
      if ( FINBitHasBeenReceived == false
         || bufferAcknowledgementNumber < sequenceNumberOfFirstFINBit )
      {
         return null ;
      }
      // A completed message is awaiting
      JNWBitSet resultBitSet = new JNWBitSet ( ) ;
      String resultString ;
      JNWBitSet singleData ;
      Boolean FINBit ;
      int i ;
      for ( i = 0 ; i < dataArray.length ; i++ )
      {
         FINBit  = ( Boolean ) dataArray [ dataArrayRowSubscript [ i ] ] [ 1 ] ;
         singleData
            = ( JNWBitSet ) dataArray [ dataArrayRowSubscript [ i ] ] [ 3 ] ;
         resultBitSet . append ( singleData ) ;
         // If this is the last packet in this message, stop
         if ( FINBit . equals ( Boolean . TRUE ) )
         {  // Is the acknowledgement number is less than the last byte in this
            // message then this messages is incomplete, so return a null
            // Record the last sequence number processed
            sequenceNumbersAlreadyProcessed 
               =  new Long ( ( Long ) ( dataArray [ dataArrayRowSubscript [ i ] ] [ 2 ] ) ) . longValue ( ) ;
            break ;
         } // end of if ( FINBit .equals ( Boolean.TRUE ) )
      }  // end of for ( i = 0 ; i < dataArray.length ; i++ )
      // Remove the packets just accumulated.  Since the array subscript begins
      // at 0, remove i + 1 rows
      RemoveBottomNRows ( i + 1 ) ;
      // remove the trailing NULLs
      resultBitSet.removeNullCharactersAtEnd ( ) ;
      // Convert to a String
      resultString = resultBitSet.convertJNWBitSetToString ( ) ;
      // First set FINBitHasBeenReceived to indicate no more messages
      resetFINBitVariables ( ) ;
      // Then look for a FIN bit for another message already in the queue
      // First reset the FIN bit variables
      for(  i = 0 ; i < dataArray.length ; i++ )
      {
         if ( ( Boolean ) dataArray [ dataArrayRowSubscript [ i ] ] [ 1 ]
            . equals ( Boolean . TRUE ) )
         {
            processFINBit ( i ) ;
            // Get the message if possible
            String subResultString = poll ( ) ;
            if ( subResultString != null )
            {
               resultString = resultString + "\n\nNEXT MESSAGE:\n\n" + subResultString ;
            }
            break ;
         }
      }  // end of for(  i = 0 ; i < dataArray.length ; i++ )
      return resultString ;
   }  // end of poll (  )


   /**Clear row n of the dataArry by moving up all the rows whose
    * index isgreater or equal to n.  This method first checks to
    * ensure that the array will not overflow.  If it will, the method
    * EnlargeDataArray ( ) is called to enlarge the array.*/
   private boolean OpenRowNOfDataArray ( int n )
   {
      int index = dataArray.length - 1 ;
      if ( n < 0 | n > index  )
      {
         Utility.err.println ("ERROR IN TCPDestinationBuffer.OpenRowIOfDataArray"
            + " attempting to open row " + n + " from "
            + " dataArray of length " + dataArray . length) ;
         return false ;
      }
      // Check the array for overflow
      if ( dataArray [ dataArrayRowSubscript [ index - 1 ] ] [ 0 ]
         != longMaxValue )
      {
         EnlargeDataArray ( ) ;
      }
      for ( int i = index ; i > n   ; i-- )
      {
         dataArray [ dataArrayRowSubscript [ i ] ] [ 0 ]
            = dataArray [ dataArrayRowSubscript [ i - 1 ] ] [ 0 ] ;
         dataArray [ dataArrayRowSubscript [ i ] ] [ 1 ]
            = dataArray [ dataArrayRowSubscript [ i - 1 ] ] [ 1 ] ;
         dataArray [ dataArrayRowSubscript [ i ] ] [ 2 ]
            = dataArray [ dataArrayRowSubscript [ i - 1 ] ] [ 2 ] ;
         dataArray [ dataArrayRowSubscript [ i ] ] [ 3 ]
            = dataArray [ dataArrayRowSubscript [ i - 1 ]] [ 3 ] ;
      }  // end of for ( int i = 0 ; i < dataArray .length ; i++ )
      return true ;
   }  // end of EnlargeDataArray


   /**Clear bottom n rows of the dataArry by moving down all the 
    * rows whose row index is n or greater.
    * @param n The number of rows to remove n > 0 and n < dataArray.length */
   private void RemoveBottomNRows ( int n )
   {
      if ( n < 1 | n > dataArray.length )
      {
         Utility.err.println ("ERROR IN TCPDestinationBuffer.RemoveBottomNRow"
            + " attempting to remove bottom " + n + " rows from "
            + " dataArray of length " + dataArray . length  ) ;
         return ;
      }
      for ( int i = n ; i < dataArray.length  ; i++ )
      {
         dataArray [ dataArrayRowSubscript [ i - n ] ] [ 0 ]
            = dataArray [ dataArrayRowSubscript [ i ] ] [ 0 ] ;
         dataArray [ dataArrayRowSubscript [ i - n ] ] [ 1 ]
            = dataArray [ dataArrayRowSubscript [ i ] ] [ 1 ] ;
         dataArray [ dataArrayRowSubscript [ i - n ] ] [ 2 ]
            = dataArray [ dataArrayRowSubscript [ i ] ] [ 2 ] ;
         dataArray [ dataArrayRowSubscript [ i - n ] ] [ 3 ]
            = dataArray [ dataArrayRowSubscript [ i ] ] [ 3 ] ;
      }  // end of for ( int i = 0 ; i < dataArray .length ; i++ )
      // Fill the newly opened row with blank data
      for ( int i = dataArray.length - n  ; i < dataArray.length   ; i++ )
      {
         dataArray [ i ] [ 0 ]  = longMaxValue ;
         dataArray [ i ] [ 1 ]  = Boolean . FALSE;
         dataArray [ i ] [ 2 ]  = longMaxValue ;       
         dataArray [ i ] [ 3 ] = null ;
      }  // end of for ( int i = 0 ; i < dataArray .length ; i++ )
   }  // end of EnlargeDataArray


   /**This checks dataArray to see if it is full.  If it is, the dataArray
    * is enlarged by five rows.
    * <p>Since the dataArray is ordered by sequence number it is not possilbe
    * to know where the new row will be inserted.  This method just
    * enlarges the array: Ir does NOT put the new row into the array.  That
    * is for the method add ( JNWBitSet IPPacketIn ) to do.
    * <p>This method is different from
    * Modeltransport.TCPState.setArrayOfTransmittedPacketInformation, which
    * enlarges the array and addes the new row to the first new row created.*/
   private void EnlargeDataArray ( )
   {
      Utility.war.println ("NOTIFICATION in Utility.TCPDestinationBuffer"
         + ".EnlargeDataArray : dataArray "
         + "overflow.\n More than " + dataArray  . length
         + " packets. Increasing dataArray size by 10."  ) ;
      // Create a new, larger dataArray
      Object [ ] [ ] newDataArray = new Object [ dataArray . length + 10 ] [ 4 ] ;
      // Copy the old array into part of the new array
      for ( int i = 0 ; i < dataArray .length ; i++ )
      {  // First copy the data array subscripts
         // Then the data
         for ( int j = 0 ; j < 4 ; j++ )
         {
            newDataArray  [ i ] [ j ]  = dataArray [ dataArrayRowSubscript [ i ] ] [ j ] ;
         }  // end of for ( int j = 0 ; j < 4 ; j++ )
      }  // end of for ( int i = 0 ; i < dataArray .length ; i++ )
      // Initialize tthe remaining part to the array to blank
      for ( int i = dataArray . length ; i < newDataArray .length ; i++ )
      {  // Set the new row values
         newDataArray  [ i ] [ 0 ] = longMaxValue ;
         newDataArray  [ i ] [ 1 ] = Boolean . FALSE;
         newDataArray  [ i ] [ 2 ] = longMaxValue ;
         newDataArray  [ i ] [ 3 ] = null ;
      }  // end of for ( int i = 0 ; i < dataArray .length ; i++ )
      dataArray =  newDataArray   ;
      // Update the subscript array
      dataArrayRowSubscript =  new int [ newDataArray  . length ] ;
      for ( int i = 0 ; i < dataArray .length ; i++ )
      {
         dataArrayRowSubscript [ i ] = i ;
      }
   }  // end of EnlargeDataArray


   /*Returns the acknowledgement number, the next sequence number
    *the queue is expecting.*/
   public long getAcknowledgementNumber ( )
   {
      return  bufferAcknowledgementNumber ;
   }


   /**This is a dummy method that is just a place holder.*/
   private void FixBufferSpace ( int decrease )
   {

   }


   /**Returns the acknowledgement number, the next sequence number
    *the TCP queue is expecting.  It starts with the first packet in the
    * dataArray and successively compares the "acknowledgement number," the
    * sequence number plus the length of data to the sequence number of
    * the next packet in the dataArray.  If they are equal then the two packets
    * are contiguous, and they can be joined.  This means that the acknowledgement
    * number is at least the "acknowledgement number" of the second packet.  The
    * method then tries to join the Nth packet with the N+1 the packet and
    * continues to repeat this to the end.*/
   private void UpdateAcknowledgementNumber ( )
   {  // If the acknowledgement number is less than the sequence
      // number for the first packet then a packet is missing
      // before the first packet in the dataArray
      //Is this test necessary?  UpdateAcknowledgementNumber ( ) is only called when
      // bufferAcknowledgementNumber == IPPacketSequenceAddress for some packet added
      // to the TCP buffer
      if ( bufferAcknowledgementNumber
         < ( ( Long ) dataArray [ dataArrayRowSubscript [ 0 ] ] [ 0 ] ) . longValue ( ) )
      {  return ;  }// Get the next sequence number expected
      long temporaryAcknowledgementNumber
         = ( ( Long ) dataArray [ dataArrayRowSubscript [ 0 ] ] [ 2 ] ) . longValue ( ) ;
      long temporarySequentialNumber ;
      for ( int i = 1 ; i < dataArray.length ; i++ )
      {
         temporarySequentialNumber
            = ( ( Long ) dataArray [ dataArrayRowSubscript [ i ]  ] [ 0 ] ).longValue ( ) ;
         if ( temporaryAcknowledgementNumber == longMaxValue )
         {  // Stop when you reach longMaxValue
            break ;  }
         if ( temporaryAcknowledgementNumber == temporarySequentialNumber )
         {
            temporaryAcknowledgementNumber
               = ( ( Long ) dataArray [ dataArrayRowSubscript [ i ]  ] [ 2 ] ) . longValue ( ) ;
         }  // end of if ( temporaryAcknowledgementNumber == temporarySequentialNumber )
         else
         { break ; }
      }  // end of for ( int i = 1 ; i < dataArray.length ; i++ )
      bufferAcknowledgementNumber = temporaryAcknowledgementNumber ;
      return ;
   }  // end of UpdateAcknowledgementNumber ( )


   /**Writes the dataArray as a String*/
   private String dataArraytoString ( )
   {
      java.lang.StringBuilder stringBuilder = new java.lang.StringBuilder ( 1000 ) ;
      // Stop at the first row that is empty
      for ( int i = 0 ; i < dataArray.length ; i++ ) {
         if ( dataArray [ dataArrayRowSubscript [ i ] ] [ 0 ]
            == longMaxValue ) {
            break ;
         }
          stringBuilder.append (
            dataArray [ dataArrayRowSubscript [ i ] ] [ 0 ] + "   "
            + dataArray [ dataArrayRowSubscript [ i ] ] [ 2 ] + "   "
            + dataArray [ dataArrayRowSubscript [ i ] ] [ 1 ] + TAB
            ) ;
          // If the data is null, append "null"
          if ( dataArray [ dataArrayRowSubscript [ i ] ] [ 3 ] == null ) {
             stringBuilder.append (
                dataArray [ dataArrayRowSubscript [ i ] ] [ 3 ]
                + CARRIAGE_RETURN ) ;
          }  // end of if
          else {  // Add the data, first changing "\n" to " "
            byte bytes [ ]
               = ((Utility.JNWBitSet ) dataArray [ dataArrayRowSubscript [ i ] ] [ 3 ])
               .convertJNWBitSetToString ().getBytes ( ) ;
            for ( int k = 0 ; k < bytes.length ; k++ ) {
               if ( bytes [ k ] == 10 ) { // from "\n"
                  bytes [ k ] = 32 ;  // to " "
               }  // end of if
            }  // end of for
            stringBuilder.append (
               new String ( bytes ) + CARRIAGE_RETURN ) ;
          }  // end of else
      }  // end of for ( int i = 0 ; i < dataArray.length ; i++ )
      return new String ( stringBuilder ) ;
   }  // end of dataArraytoString


   /** For testing only*/
   public static void main_For_Testing_Only( String [ ] args )
   {
   Long longMaxValue = new Long ( ( long) 1000 * ( long) 1000 * 1000 * 100 -1 ) ;
      System.out.println (" longMaxValue = " + longMaxValue ) ;
      Long TWO = new Long ( ( long ) 1024 * 1024 * 1024 * 32 ) ;
      System.out.println (" TWO = " + TWO  ) ;
      System.exit ( 1 ) ;
      /*
      JNWBitSet test3 = new JNWBitSet ( ) ;
      test3.setJNWBitSetToString ( "A" ) ;
      JNWBitSet clone = new JNWBitSet ( 0 );
      for ( int i = 0 ; i < 7 ; i++ ) {
         clone = test3 . clone ( ) ;
         clone . shift ( 8 * i ) ;
         clone = RemoveNullCharactersAtEnd ( clone ) ;
         System.out.println ( " i = " + i + " message = :" + clone.convertJNWBitSetToString () +":" ) ;
      }

      System.exit ( 0 ) ;
       */

  /*
      TCPDestinationBuffer buffer = new TCPDestinationBuffer ( ) ;
      JNWBitSet IPPacket  ;
      JNWBitSet IPPacketHeader  ;
      JNWBitSet TCPHeader  ;
      long sequenceNumber211 ;
      long acknowledgementNumber ;
      int dataLength ;
      long sequenceRange = 1000000 ;
      int TCPSourcePort = 23 ;
      int TCPDestinationPort = TCPSourcePort  ;
      int messageLengthInBytes ;
      int windowSize = 256 ;
      Utility.JNWRandom randomNumberGenerator = Utility.JNWRandom.getInstance () ;
      randomNumberGenerator.setSeed ( 9872741129348L ) ;




      for ( int i = 1 ; i <= 77 ; i++ )
      {
         IPPacket = new JNWBitSet ( 20 * 8 ) ;
         TCPHeader = new JNWBitSet  ( ) ;
         messageLengthInBytes = randomNumberGenerator.nextInt ( 100 ) ;
         messageLengthInBytes = 8 ;
         // Now construct the TCPHeader
         JNWBitSet sourcePortBitSet = new JNWBitSet ( TCPSourcePort , 16  ) ;
         // Assumes that we are sending between ports 2
         JNWBitSet destinationPortBitSet = new JNWBitSet ( TCPDestinationPort , 16 ) ;

         sequenceNumber211 = ( long ) randomNumberGenerator . nextInt ( ( int ) sequenceRange ) ;

         System.out.println ( " sequence number = " + sequenceNumber211 ) ;
         acknowledgementNumber = sequenceNumber211 + messageLengthInBytes ;
         JNWBitSet sequenceNumberBitSet = new JNWBitSet ( sequenceNumber211 , 32 ) ;
         JNWBitSet acknowledgementNumberBitSet = new JNWBitSet ( acknowledgementNumber , 32 ) ;
         JNWBitSet TCPHeaderLength = new JNWBitSet ( 5 , 4 ) ;
         JNWBitSet sixZeros = new JNWBitSet ( "000000" ) ;
         JNWBitSet flags = new JNWBitSet ( "000000" ) ;
         JNWBitSet windowSizeBitSet = new JNWBitSet ( windowSize , 16  ) ;
         // Sixteen zeros whihc will changed later
         JNWBitSet checksum = new JNWBitSet ( 16 );
         // Sixteen zeros
         JNWBitSet urgentPointer = new JNWBitSet ( 16 );
         TCPHeader . prepend ( urgentPointer ) ;
         TCPHeader . prepend ( checksum ) ;
         TCPHeader . prepend ( windowSizeBitSet ) ;
         TCPHeader . prepend ( flags) ;
         TCPHeader . prepend ( sixZeros ) ;
         TCPHeader . prepend ( TCPHeaderLength ) ;
         TCPHeader . prepend ( acknowledgementNumberBitSet ) ;
         TCPHeader . prepend ( sequenceNumberBitSet ) ;
         TCPHeader . prepend ( destinationPortBitSet ) ;
         TCPHeader . prepend ( sourcePortBitSet ) ;
         TCPHeader . append ( new JNWBitSet ( BITS_PER_BYTE * messageLengthInBytes)) ;
         IPPacket . append ( TCPHeader ) ;
         buffer.add ( IPPacket  )  ;
      }
      System.out.println ( "\n\n\n\n\n\n" ) ;
      System.out.println ( "dataArray = :\n" + buffer . dataArraytoString (   ) ) ;
      ////   for ( int i = 1 ; i <= 15 ; i = i + 2 ) {
      buffer . RemoveBottomNRows ( 70 ) ;
      System.out.println ( "dataArray = :\n" + buffer . dataArraytoString (   ) ) ;
      ////   }
      System.exit ( 0 ) ;




      for ( int i = 0 ; i < buffer.dataArray.length + 2 ; i++ )
      {
         buffer = new TCPDestinationBuffer ( ) ;
         buffer.RemoveBottomNRows ( i ) ;
         System.out.println ( "dataArray for " + i + " = :\n" + buffer . dataArraytoString (   ) ) ;
      }

      java.lang.Long number1 = ( java.lang.Long ) 1234341234L ;
      java.lang.Long number2 = ( java.lang.Long ) 987689769L ;
      /////      if ( number1.)


      System.exit ( 0 ) ;
   */

      System.out.println ( " Starting TCPJNWBiffer test " ) ;
      String emailMessage = "Now it the time for all good men to come to the aid of their country." ;
      JNWBitSet eMailBitSet = new JNWBitSet ( ) ;
      eMailBitSet . setJNWBitSetToString ( emailMessage ) ;
      TCPDestinationBuffer destinationBuffer = new TCPDestinationBuffer ( ) ;
      Utility.JNWBitSet TCPPackets [] ;
      ModelTransport . TCPConnection tcpConnection  = new ModelTransport . TCPConnection ( 0 , 5 , 10) ;
      tcpConnection.maximumTCPPayloadInBytes = 5 ;

      String emailMessage1 = "0000011111222223333344444555556666677777888889999900000" ;
      JNWBitSet eMailBitSet1 = new JNWBitSet ( ) ;
      eMailBitSet1 . setJNWBitSetToString ( emailMessage1 ) ;

      String emailMessage2 = "aaaaabbbbbcccccdddddeeeeefffffggggghhhhhiiiiijjjjjkkkkk" ;
      JNWBitSet eMailBitSet2 = new JNWBitSet ( ) ;
      eMailBitSet2 . setJNWBitSetToString ( emailMessage2 ) ;

      String emailMessage3 = "lllllmmmmmnnnnnooooopppppqqqqqrrrrrssssstttttu" ;
      JNWBitSet eMailBitSet3 = new JNWBitSet ( ) ;
      eMailBitSet3 . setJNWBitSetToString ( emailMessage3 ) ;

      Utility.JNWBitSet TCPPackets1 [] ;
      Utility.JNWBitSet TCPPackets2 [] ;
      Utility.JNWBitSet TCPPackets3 [] ;

      TCPPackets1  = tcpConnection . PrependTCPHeadersToData ( eMailBitSet1
         , 0 , 5 ) ;
      TCPPackets2  = tcpConnection . PrependTCPHeadersToData ( eMailBitSet2
         , 0 , 5 ) ;
      ////System.out.println ( " TCPPackets2 [ 0 ] =" + TCPPackets2 [ 0 ].writeBitSet ()) ;
      TCPPackets3  = tcpConnection . PrependTCPHeadersToData ( eMailBitSet3
         , 0 , 5 ) ;

      for ( int i = 0 ; i < TCPPackets1.length ; i++ )
      {
         TCPPackets1 [ i ]
            =  ModelTransport.IPUtilities.PrependIPHeaderToMessage (  TCPPackets1 [ i ]
            , 0 , 5 , ModelData.GlobalVariables.transmissionProtocols.TCP)  ;
      }  // end of for ( int i = 0 ; i < TCPPackets.length ; i++ )


      for ( int i = 0 ; i < TCPPackets2.length ; i++ )
      {
         TCPPackets2 [ i ]
            =  ModelTransport.IPUtilities.PrependIPHeaderToMessage (  TCPPackets2 [ i ]
            , 0 , 5 , ModelData.GlobalVariables.transmissionProtocols.TCP)  ;
      }  // end of for ( int i = 0 ; i < TCPPackets.length ; i++ )

      for ( int i = 0 ; i < TCPPackets2.length ; i++ )
      {
         destinationBuffer . add ( TCPPackets2 [  i  ]  . clone ( ) ) ;
      }  // end of for ( int i = 0 ; i < TCPPackets.length ; i++ )

      for ( int i = 0 ; i < TCPPackets3.length  ; i++ )
      {
         TCPPackets3 [ i ]
            =  ModelTransport.IPUtilities.PrependIPHeaderToMessage (  TCPPackets3 [ i ]
            , 0 , 5 , ModelData.GlobalVariables.transmissionProtocols.TCP)  ;

      }  // end of for ( int i = 0 ; i < TCPPackets.length ; i++ )

      for ( int i = 0 ; i < TCPPackets3.length  ; i++ )
      {
         destinationBuffer . add ( TCPPackets3 [  i  ]  . clone ( ) ) ;

      }  // end of for ( int i = 0 ; i < TCPPackets.length ; i++ )


      String result = destinationBuffer . poll ( )  ;
      if ( result != null )
      {
         System.out.println ("Message as String = \n" +  result   )   ;
      }
      else
      {
         System.out.println ("No complete message "  )   ;
      }
      for ( int i = 0 ; i < TCPPackets1.length ; i++ )
      {
         destinationBuffer . add ( TCPPackets1 [  i  ]  . clone ( ) ) ;
         ////        destinationBuffer . UpdateAcknowledgementNumber () ;
      }  // end of for ( int i = 0 ; i < TCPPackets.length ; i++ )

      result = destinationBuffer . poll ( )  ;
      if ( result != null )
      {
         System.out.println ("Message as String = \n" +  result   )   ;
      }
      else
      {
         System.out.println ("No complete message "  )   ;
      }

   } // end of main




}  // end of public class TCPDestinationBuffer